import einops as ein
import torch
from torch import Tensor, nn
from torchtyping import TensorType


class RMSNorm(nn.Module):
    """
    Implements Root Mean Square Normalization introduced in
    https://arxiv.org/pdf/1910.07467.pdf.

    Reference implementation (used for correctness verfication)
    can be found here:
    https://github.com/facebookresearch/llama/blob/main/llama/model.py

    Args:
        dim (int): embedding size
        eps (float): small value to avoid division by zero. Default: 1e-6
    """

    def __init__(self, dim: int, eps: float = 1e-6) -> None:
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x (Tensor): input tensor to normalize

        Returns:
            Tensor: The output tensor after applying RMSNorm.
        """
        # computation is in fp32
        x_fp32 = x.float()
        x_normed = (
            x_fp32 * torch.rsqrt(x_fp32.pow(2).mean(-1, keepdim=True) + self.eps)
        ).type_as(x)
        return x_normed * self.scale


class ReLUKANLayer(nn.Module):
    def __init__(
        self, input_size: int, g: int, k: int, output_size: int, train_ab: bool = True
    ):
        super().__init__()
        self.g, self.k, self.r = g, k, 4 * g * g / ((k + 1) * (k + 1))
        self.input_size, self.output_size = input_size, output_size

        phase_low = torch.arange(-k, g, dtype=torch.float32) / g
        phase_height = phase_low + (k + 1) / g
        self.phase_low = nn.Parameter(
            phase_low.repeat(input_size, 1),
            requires_grad=train_ab,
        )
        self.phase_height = nn.Parameter(
            phase_height.repeat(input_size, 1),
            requires_grad=train_ab,
        )

        self.equal_size_conv = nn.Conv2d(1, output_size, (g + k, input_size))

    def forward(
        self, x: TensorType["B", "input_size"]
    ) -> TensorType["B", "output_size"]:
        x = x.unsqueeze(-1)
        x1 = F.relu(x - self.phase_low)
        x2 = F.relu(self.phase_height - x)
        x = x1 * x2 * self.r
        x = x * x
        x = ein.rearrange(x, "... in gk -> ... 1 gk in")
        x = self.equal_size_conv(x)
        x = ein.rearrange(x, "... out 1 1 -> ... out")
        return x



import torch
from torch import nn


class LinearNoiseSchedule(nn.Module):
    def __init__(self, beta_start=1e-4, beta_end=0.02, num_timesteps=1000):
        super().__init__()
        self.num_timesteps = num_timesteps

        betas = torch.linspace(beta_start, beta_end, num_timesteps)
        alphas = 1 - betas
        alphas_cumprod = torch.cumprod(alphas, dim=0)

        self.register_buffer("betas", betas)
        self.register_buffer("alphas", alphas)
        self.register_buffer("alphas_cumprod", alphas_cumprod)

    def get_alpha(self, t):
        return self.alphas[t]

    def get_variance(self, t):
        return 1 - self.alphas[t]

    def get_previous_alpha(self, t):
        return self.alphas_cumprod[t - 1] if t > 0 else 1

    def sample_noise(self, x_0, t):
        noise = torch.randn_like(x_0)
        sqrt_alpha_cumprod = torch.sqrt(self.alphas_cumprod[t])
        sqrt_one_minus_alpha_cumprod = torch.sqrt(1 - self.alphas_cumprod[t])
        return sqrt_alpha_cumprod * x_0 + sqrt_one_minus_alpha_cumprod * noise

    def sample_noise_from_output(self, x_0, t, noise_pred):
        sqrt_one_minus_alpha_cumprod = torch.sqrt(1 - self.alphas_cumprod[t])
        return x_0 - sqrt_one_minus_alpha_cumprod * noise_pred


class CosineNoiseSchedule(nn.Module):
    def __init__(self, max_beta=0.999, noise_cosine_shift=0.008, num_timesteps=1000):
        super().__init__()
        self.num_timesteps = num_timesteps

        timesteps = torch.arange(1, num_timesteps + 1, dtype=torch.float64)
        alphas_cumprod = self._cosine_variance_schedule(timesteps / num_timesteps)
        betas = 1 - alphas_cumprod[1:] / alphas_cumprod[:-1]

        self.register_buffer("betas", betas)
        self.register_buffer("alphas_cumprod", alphas_cumprod)
        self.noise_cosine_shift = noise_cosine_shift
        self.max_beta = max_beta

    def get_alpha(self, t):
        return self.alphas_cumprod[t + 1] / self.alphas_cumprod[t]

    def get_variance(self, t):
        return self.betas[t]

    def get_previous_alpha(self, t):
        return self.alphas_cumprod[t - 1] if t > 0 else 1

    def sample_noise(self, x_0, t):
        noise = torch.randn_like(x_0)
        sqrt_alpha_cumprod = torch.sqrt(self.alphas_cumprod[t])
        sqrt_one_minus_alpha_cumprod = torch.sqrt(1 - self.alphas_cumprod[t])
        return sqrt_alpha_cumprod * x_0 + sqrt_one_minus_alpha_cumprod * noise

    def sample_noise_from_output(self, x_0, t, noise_pred):
        sqrt_one_minus_alpha_cumprod = torch.sqrt(1 - self.alphas_cumprod[t])
        return x_0 - sqrt_one_minus_alpha_cumprod * noise_pred

    def _cosine_variance_schedule(self, t):
        s = self.noise_cosine_shift
        max_beta = self.max_beta
        f_t = torch.cos((t + s) / (1 + s) * torch.pi / 2) ** 2
        f_0 = torch.cos(s / (1 + s) * torch.pi / 2) ** 2
        return f_0 - (f_0 - max_beta) * (f_t / f_0)



import math
import warnings

import torch
import torch.distributions as D
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.overrides import handle_torch_function, has_torch_function_variadic
import torch
import torch.distributions as D
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


def tanh_exp(x, threshold=3.0):
    """
    TanhExp(x) = x * tanh(exp(x))

    - Clamp is necessary to prevent overflow. Using torch.where alone is insufficient;
        there might be issues when x is small.

    - TanhExp converges to 1 when x is large;  x*tanh(exp(x)) - x < 0f64 if x > 3
    """
    return torch.where(
        x > threshold, x, x * torch.tanh(torch.exp(torch.clamp(x, max=threshold)))
    )


class TanhExp(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, threshold=3.0):
        return tanh_exp(x, threshold)


def quantile_loss(
    input: Tensor,
    target: Tensor,
    quantile: float,
    reduction: str = "mean",
) -> Tensor:
    r"""quantile_loss(input, target, quantile, reduction='mean') -> Tensor

    Measures the element-wise quantile loss.

    The quantile loss is defined as:

    .. math::
        \begin{aligned}
        \ell(x, y) = \begin{cases}
            (1 - q) * (y - x), & \text{if } x < y \\
            q * (x - y), & \text{otherwise }
        \end{cases}
        \end{aligned}

    where :math:`x` is the input tensor, :math:`y` is the target tensor, and :math:`q` is the quantile.

    Args:
        input (Tensor): the input tensor
        target (Tensor): the target tensor
        quantile (float): the quantile value, should be between 0 and 1
        reduction (str, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed. Default: ``'mean'``

    Returns:
        Tensor: the quantile loss
    """
    if has_torch_function_variadic(input, target):
        return handle_torch_function(
            quantile_loss,
            (input, target),
            input,
            target,
            quantile=quantile,
            reduction=reduction,
        )
    if not (target.size() == input.size()):
        warnings.warn(
            f"Using a target size ({target.size()}) that is different to the input size ({input.size()}). "
            "This will likely lead to incorrect results due to broadcasting. "
            "Please ensure they have the same size.",
            stacklevel=2,
        )
    if not (0 <= quantile <= 1):
        raise ValueError(f"Quantile value should be between 0 and 1, got {quantile}")

    expanded_input, expanded_target = torch.broadcast_tensors(input, target)
    diff = expanded_target - expanded_input
    loss = torch.max((quantile - 1) * diff, quantile * diff)

    if reduction == "mean":
        loss = torch.mean(loss)
    elif reduction == "sum":
        loss = torch.sum(loss)
    elif reduction != "none":
        raise ValueError(f"Invalid reduction: {reduction}")

    return loss


class QuantileLoss(nn.Module):
    def __init__(self, quantile: float, reduction: str = "mean"):
        super().__init__()
        self.quantile = quantile
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return quantile_loss(input, target, self.quantile, self.reduction)



# Ref. https://github.com/pytorch/rl/blob/3a7cf6af2a08089f11e0ed8cad3dd1cea0e253fb/torchrl/data/datasets/d4rl_infos.py#L7
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

DATASETS_URLS = {
    "maze2d-open-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-open-sparse.hdf5",
    "maze2d-umaze-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-umaze-sparse-v1.hdf5",
    "maze2d-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-medium-sparse-v1.hdf5",
    "maze2d-large-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-large-sparse-v1.hdf5",
    "maze2d-eval-umaze-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-umaze-sparse-v1.hdf5",
    "maze2d-eval-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-medium-sparse-v1.hdf5",
    "maze2d-eval-large-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-large-sparse-v1.hdf5",
    "maze2d-open-dense-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-open-dense.hdf5",
    "maze2d-umaze-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-umaze-dense-v1.hdf5",
    "maze2d-medium-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-medium-dense-v1.hdf5",
    "maze2d-large-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-large-dense-v1.hdf5",
    "maze2d-eval-umaze-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-umaze-dense-v1.hdf5",
    "maze2d-eval-medium-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-medium-dense-v1.hdf5",
    "maze2d-eval-large-dense-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-eval-large-dense-v1.hdf5",
    "minigrid-fourrooms-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/minigrid/minigrid4rooms.hdf5",
    "minigrid-fourrooms-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/minigrid/minigrid4rooms_random.hdf5",
    "pen-human-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/pen-v0_demos_clipped.hdf5",
    "pen-cloned-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/pen-demos-v0-bc-combined.hdf5",
    "pen-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/pen-v0_expert_clipped.hdf5",
    "hammer-human-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/hammer-v0_demos_clipped.hdf5",
    "hammer-cloned-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/hammer-demos-v0-bc-combined.hdf5",
    "hammer-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/hammer-v0_expert_clipped.hdf5",
    "relocate-human-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/relocate-v0_demos_clipped.hdf5",
    "relocate-cloned-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/relocate-demos-v0-bc-combined.hdf5",
    "relocate-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/relocate-v0_expert_clipped.hdf5",
    "door-human-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/door-v0_demos_clipped.hdf5",
    "door-cloned-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/door-demos-v0-bc-combined.hdf5",
    "door-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/door-v0_expert_clipped.hdf5",
    "halfcheetah-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/halfcheetah_random.hdf5",
    "halfcheetah-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/halfcheetah_medium.hdf5",
    "halfcheetah-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/halfcheetah_expert.hdf5",
    "halfcheetah-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/halfcheetah_mixed.hdf5",
    "halfcheetah-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/halfcheetah_medium_expert.hdf5",
    "walker2d-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/walker2d_random.hdf5",
    "walker2d-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/walker2d_medium.hdf5",
    "walker2d-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/walker2d_expert.hdf5",
    "walker2d-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/walker_mixed.hdf5",
    "walker2d-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/walker2d_medium_expert.hdf5",
    "hopper-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/hopper_random.hdf5",
    "hopper-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/hopper_medium.hdf5",
    "hopper-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/hopper_expert.hdf5",
    "hopper-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/hopper_mixed.hdf5",
    "hopper-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/hopper_medium_expert.hdf5",
    "ant-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_random.hdf5",
    "ant-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_medium.hdf5",
    "ant-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_expert.hdf5",
    "ant-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_mixed.hdf5",
    "ant-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_medium_expert.hdf5",
    "ant-random-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco/ant_random_expert.hdf5",
    "antmaze-umaze-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_u-maze_noisy_multistart_False_multigoal_False_sparse.hdf5",
    "antmaze-umaze-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_u-maze_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-medium-play-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_big-maze_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-medium-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_big-maze_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-large-play-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_hardest-maze_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-large-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_hardest-maze_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-umaze-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_u-maze_noisy_multistart_False_multigoal_False_sparse_fixed.hdf5",
    "antmaze-umaze-diverse-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_u-maze_noisy_multistart_True_multigoal_True_sparse_fixed.hdf5",
    "antmaze-medium-play-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_big-maze_noisy_multistart_True_multigoal_False_sparse_fixed.hdf5",
    "antmaze-medium-diverse-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_big-maze_noisy_multistart_True_multigoal_True_sparse_fixed.hdf5",
    "antmaze-large-play-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_hardest-maze_noisy_multistart_True_multigoal_False_sparse_fixed.hdf5",
    "antmaze-large-diverse-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v2/Ant_maze_hardest-maze_noisy_multistart_True_multigoal_True_sparse_fixed.hdf5",
    "flow-ring-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/flow/flow-ring-v0-random.hdf5",
    "flow-ring-controller-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/flow/flow-ring-v0-idm.hdf5",
    "flow-merge-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/flow/flow-merge-v0-random.hdf5",
    "flow-merge-controller-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/flow/flow-merge-v0-idm.hdf5",
    "kitchen-complete-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/kitchen/mini_kitchen_microwave_kettle_light_slider-v0.hdf5",
    "kitchen-partial-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/kitchen/kitchen_microwave_kettle_light_slider-v0.hdf5",
    "kitchen-mixed-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/kitchen/kitchen_microwave_kettle_bottomburner_light-v0.hdf5",
    "carla-lane-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/carla/carla_lane_follow_flat-v0.hdf5",
    "carla-town-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/carla/carla_town_subsamp_flat-v0.hdf5",
    "carla-town-full-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/carla/carla_town_flat-v0.hdf5",
    "bullet-halfcheetah-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-halfcheetah_random.hdf5",
    "bullet-halfcheetah-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-halfcheetah_medium.hdf5",
    "bullet-halfcheetah-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-halfcheetah_expert.hdf5",
    "bullet-halfcheetah-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-halfcheetah_medium_expert.hdf5",
    "bullet-halfcheetah-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-halfcheetah_medium_replay.hdf5",
    "bullet-hopper-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-hopper_random.hdf5",
    "bullet-hopper-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-hopper_medium.hdf5",
    "bullet-hopper-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-hopper_expert.hdf5",
    "bullet-hopper-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-hopper_medium_expert.hdf5",
    "bullet-hopper-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-hopper_medium_replay.hdf5",
    "bullet-ant-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-ant_random.hdf5",
    "bullet-ant-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-ant_medium.hdf5",
    "bullet-ant-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-ant_expert.hdf5",
    "bullet-ant-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-ant_medium_expert.hdf5",
    "bullet-ant-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-ant_medium_replay.hdf5",
    "bullet-walker2d-random-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-walker2d_random.hdf5",
    "bullet-walker2d-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-walker2d_medium.hdf5",
    "bullet-walker2d-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-walker2d_expert.hdf5",
    "bullet-walker2d-medium-expert-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-walker2d_medium_expert.hdf5",
    "bullet-walker2d-medium-replay-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-walker2d_medium_replay.hdf5",
    "bullet-maze2d-open-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-maze2d-open-sparse.hdf5",
    "bullet-maze2d-umaze-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-maze2d-umaze-sparse.hdf5",
    "bullet-maze2d-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-maze2d-medium-sparse.hdf5",
    "bullet-maze2d-large-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/bullet/bullet-maze2d-large-sparse.hdf5",
    "halfcheetah-random-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_random-v1.hdf5",
    "halfcheetah-random-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_random-v2.hdf5",
    "halfcheetah-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_medium-v1.hdf5",
    "halfcheetah-medium-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_medium-v2.hdf5",
    "halfcheetah-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_expert-v1.hdf5",
    "halfcheetah-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_expert-v2.hdf5",
    "halfcheetah-medium-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_medium_replay-v1.hdf5",
    "halfcheetah-medium-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_medium_replay-v2.hdf5",
    "halfcheetah-full-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_full_replay-v1.hdf5",
    "halfcheetah-full-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_full_replay-v2.hdf5",
    "halfcheetah-medium-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/halfcheetah_medium_expert-v1.hdf5",
    "halfcheetah-medium-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/halfcheetah_medium_expert-v2.hdf5",
    "hopper-random-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_random-v1.hdf5",
    "hopper-random-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_random-v2.hdf5",
    "hopper-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_medium-v1.hdf5",
    "hopper-medium-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_medium-v2.hdf5",
    "hopper-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_expert-v1.hdf5",
    "hopper-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_expert-v2.hdf5",
    "hopper-medium-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_medium_replay-v1.hdf5",
    "hopper-medium-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_medium_replay-v2.hdf5",
    "hopper-full-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_full_replay-v1.hdf5",
    "hopper-full-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_full_replay-v2.hdf5",
    "hopper-medium-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/hopper_medium_expert-v1.hdf5",
    "hopper-medium-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/hopper_medium_expert-v2.hdf5",
    "walker2d-random-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_random-v1.hdf5",
    "walker2d-random-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_random-v2.hdf5",
    "walker2d-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_medium-v1.hdf5",
    "walker2d-medium-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_medium-v2.hdf5",
    "walker2d-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_expert-v1.hdf5",
    "walker2d-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_expert-v2.hdf5",
    "walker2d-medium-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_medium_replay-v1.hdf5",
    "walker2d-medium-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_medium_replay-v2.hdf5",
    "walker2d-full-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_full_replay-v1.hdf5",
    "walker2d-full-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_full_replay-v2.hdf5",
    "walker2d-medium-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/walker2d_medium_expert-v1.hdf5",
    "walker2d-medium-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/walker2d_medium_expert-v2.hdf5",
    "ant-random-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_random-v1.hdf5",
    "ant-random-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_random-v2.hdf5",
    "ant-medium-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_medium-v1.hdf5",
    "ant-medium-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_medium-v2.hdf5",
    "ant-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_expert-v1.hdf5",
    "ant-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_expert-v2.hdf5",
    "ant-medium-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_medium_replay-v1.hdf5",
    "ant-medium-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_medium_replay-v2.hdf5",
    "ant-full-replay-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_full_replay-v1.hdf5",
    "ant-full-replay-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_full_replay-v2.hdf5",
    "ant-medium-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v1/ant_medium_expert-v1.hdf5",
    "ant-medium-expert-v2": "http://rail.eecs.berkeley.edu/datasets/offline_rl/gym_mujoco_v2/ant_medium_expert-v2.hdf5",
    "hammer-human-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/hammer-human-v1.hdf5",
    "hammer-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/hammer-expert-v1.hdf5",
    "hammer-cloned-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/hammer-cloned-v1.hdf5",
    "pen-human-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/pen-human-v1.hdf5",
    "pen-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/pen-expert-v1.hdf5",
    "pen-cloned-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/pen-cloned-v1.hdf5",
    "relocate-human-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/relocate-human-v1.hdf5",
    "relocate-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/relocate-expert-v1.hdf5",
    "relocate-cloned-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/relocate-cloned-v1.hdf5",
    "door-human-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/door-human-v1.hdf5",
    "door-expert-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/door-expert-v1.hdf5",
    "door-cloned-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg_v1/door-cloned-v1.hdf5",
    "antmaze-umaze-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_umaze_noisy_multistart_False_multigoal_False_sparse.hdf5",
    "antmaze-umaze-diverse-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_umaze_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-medium-play-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_medium_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-medium-diverse-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_medium_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-large-diverse-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_large_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-large-play-v1": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_v1/Ant_maze_large_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-eval-umaze-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_umaze_eval_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-eval-umaze-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_umaze_eval_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-eval-medium-play-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_medium_eval_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "antmaze-eval-medium-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_medium_eval_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-eval-large-diverse-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_large_eval_noisy_multistart_True_multigoal_False_sparse.hdf5",
    "antmaze-eval-large-play-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/ant_maze_new/Ant_maze_large_eval_noisy_multistart_True_multigoal_True_sparse.hdf5",
    "door-human-longhorizon-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/door-v0_demos_clipped.hdf5",
    "hammer-human-longhorizon-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/hammer-v0_demos_clipped.hdf5",
    "pen-human-longhorizon-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/pen-v0_demos_clipped.hdf5",
    "relocate-human-longhorizon-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/hand_dapg/relocate-v0_demos_clipped.hdf5",
    "maze2d-umaze-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-umaze-sparse.hdf5",
    "maze2d-medium-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-medium-sparse.hdf5",
    "maze2d-large-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-large-sparse.hdf5",
    "maze2d-umaze-dense-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-umaze-dense.hdf5",
    "maze2d-medium-dense-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-medium-dense.hdf5",
    "maze2d-large-dense-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/maze2d/maze2d-large-dense.hdf5",
    "carla-lane-render-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/carla/carla_lane_follow-v0.hdf5",
    "carla-town-render-v0": "http://rail.eecs.berkeley.edu/datasets/offline_rl/carla/carla_town_flat-v0.hdf5",
}


REF_MIN_SCORE = {
    "maze2d-open-v0": 0.01,
    "maze2d-umaze-v1": 23.85,
    "maze2d-medium-v1": 13.13,
    "maze2d-large-v1": 6.7,
    "maze2d-open-dense-v0": 11.17817,
    "maze2d-umaze-dense-v1": 68.537689,
    "maze2d-medium-dense-v1": 44.264742,
    "maze2d-large-dense-v1": 30.569041,
    "minigrid-fourrooms-v0": 0.01442,
    "minigrid-fourrooms-random-v0": 0.01442,
    "pen-human-v0": 96.262799,
    "pen-cloned-v0": 96.262799,
    "pen-expert-v0": 96.262799,
    "hammer-human-v0": -274.856578,
    "hammer-cloned-v0": -274.856578,
    "hammer-expert-v0": -274.856578,
    "relocate-human-v0": -6.425911,
    "relocate-cloned-v0": -6.425911,
    "relocate-expert-v0": -6.425911,
    "door-human-v0": -56.512833,
    "door-cloned-v0": -56.512833,
    "door-expert-v0": -56.512833,
    "halfcheetah-random-v0": -280.178953,
    "halfcheetah-medium-v0": -280.178953,
    "halfcheetah-expert-v0": -280.178953,
    "halfcheetah-medium-replay-v0": -280.178953,
    "halfcheetah-medium-expert-v0": -280.178953,
    "walker2d-random-v0": 1.629008,
    "walker2d-medium-v0": 1.629008,
    "walker2d-expert-v0": 1.629008,
    "walker2d-medium-replay-v0": 1.629008,
    "walker2d-medium-expert-v0": 1.629008,
    "hopper-random-v0": -20.272305,
    "hopper-medium-v0": -20.272305,
    "hopper-expert-v0": -20.272305,
    "hopper-medium-replay-v0": -20.272305,
    "hopper-medium-expert-v0": -20.272305,
    "ant-random-v0": -325.6,
    "ant-medium-v0": -325.6,
    "ant-expert-v0": -325.6,
    "ant-medium-replay-v0": -325.6,
    "ant-medium-expert-v0": -325.6,
    "antmaze-umaze-v0": 0.0,
    "antmaze-umaze-diverse-v0": 0.0,
    "antmaze-medium-play-v0": 0.0,
    "antmaze-medium-diverse-v0": 0.0,
    "antmaze-large-play-v0": 0.0,
    "antmaze-large-diverse-v0": 0.0,
    "antmaze-umaze-v2": 0.0,
    "antmaze-umaze-diverse-v2": 0.0,
    "antmaze-medium-play-v2": 0.0,
    "antmaze-medium-diverse-v2": 0.0,
    "antmaze-large-play-v2": 0.0,
    "antmaze-large-diverse-v2": 0.0,
    "kitchen-complete-v0": 0.0,
    "kitchen-partial-v0": 0.0,
    "kitchen-mixed-v0": 0.0,
    "flow-ring-random-v0": -165.22,
    "flow-ring-controller-v0": -165.22,
    "flow-merge-random-v0": 118.67993,
    "flow-merge-controller-v0": 118.67993,
    "carla-lane-v0": -0.8503839912088142,
    "carla-town-v0": -114.81579500772153,  # random score
    "bullet-halfcheetah-random-v0": -1275.766996,
    "bullet-halfcheetah-medium-v0": -1275.766996,
    "bullet-halfcheetah-expert-v0": -1275.766996,
    "bullet-halfcheetah-medium-expert-v0": -1275.766996,
    "bullet-halfcheetah-medium-replay-v0": -1275.766996,
    "bullet-hopper-random-v0": 20.058972,
    "bullet-hopper-medium-v0": 20.058972,
    "bullet-hopper-expert-v0": 20.058972,
    "bullet-hopper-medium-expert-v0": 20.058972,
    "bullet-hopper-medium-replay-v0": 20.058972,
    "bullet-ant-random-v0": 373.705955,
    "bullet-ant-medium-v0": 373.705955,
    "bullet-ant-expert-v0": 373.705955,
    "bullet-ant-medium-expert-v0": 373.705955,
    "bullet-ant-medium-replay-v0": 373.705955,
    "bullet-walker2d-random-v0": 16.523877,
    "bullet-walker2d-medium-v0": 16.523877,
    "bullet-walker2d-expert-v0": 16.523877,
    "bullet-walker2d-medium-expert-v0": 16.523877,
    "bullet-walker2d-medium-replay-v0": 16.523877,
    "bullet-maze2d-open-v0": 8.750000,
    "bullet-maze2d-umaze-v0": 32.460000,
    "bullet-maze2d-medium-v0": 14.870000,
    "bullet-maze2d-large-v0": 1.820000,
}

REF_MAX_SCORE = {
    "maze2d-open-v0": 20.66,
    "maze2d-umaze-v1": 161.86,
    "maze2d-medium-v1": 277.39,
    "maze2d-large-v1": 273.99,
    "maze2d-open-dense-v0": 27.166538620695782,
    "maze2d-umaze-dense-v1": 193.66285642381482,
    "maze2d-medium-dense-v1": 297.4552547777125,
    "maze2d-large-dense-v1": 303.4857382709002,
    "minigrid-fourrooms-v0": 2.89685,
    "minigrid-fourrooms-random-v0": 2.89685,
    "pen-human-v0": 3076.8331017826877,
    "pen-cloned-v0": 3076.8331017826877,
    "pen-expert-v0": 3076.8331017826877,
    "hammer-human-v0": 12794.134825156867,
    "hammer-cloned-v0": 12794.134825156867,
    "hammer-expert-v0": 12794.134825156867,
    "relocate-human-v0": 4233.877797728884,
    "relocate-cloned-v0": 4233.877797728884,
    "relocate-expert-v0": 4233.877797728884,
    "door-human-v0": 2880.5693087298737,
    "door-cloned-v0": 2880.5693087298737,
    "door-expert-v0": 2880.5693087298737,
    "halfcheetah-random-v0": 12135.0,
    "halfcheetah-medium-v0": 12135.0,
    "halfcheetah-expert-v0": 12135.0,
    "halfcheetah-medium-replay-v0": 12135.0,
    "halfcheetah-medium-expert-v0": 12135.0,
    "walker2d-random-v0": 4592.3,
    "walker2d-medium-v0": 4592.3,
    "walker2d-expert-v0": 4592.3,
    "walker2d-medium-replay-v0": 4592.3,
    "walker2d-medium-expert-v0": 4592.3,
    "hopper-random-v0": 3234.3,
    "hopper-medium-v0": 3234.3,
    "hopper-expert-v0": 3234.3,
    "hopper-medium-replay-v0": 3234.3,
    "hopper-medium-expert-v0": 3234.3,
    "ant-random-v0": 3879.7,
    "ant-medium-v0": 3879.7,
    "ant-expert-v0": 3879.7,
    "ant-medium-replay-v0": 3879.7,
    "ant-medium-expert-v0": 3879.7,
    "antmaze-umaze-v0": 1.0,
    "antmaze-umaze-diverse-v0": 1.0,
    "antmaze-medium-play-v0": 1.0,
    "antmaze-medium-diverse-v0": 1.0,
    "antmaze-large-play-v0": 1.0,
    "antmaze-large-diverse-v0": 1.0,
    "antmaze-umaze-v2": 1.0,
    "antmaze-umaze-diverse-v2": 1.0,
    "antmaze-medium-play-v2": 1.0,
    "antmaze-medium-diverse-v2": 1.0,
    "antmaze-large-play-v2": 1.0,
    "antmaze-large-diverse-v2": 1.0,
    "kitchen-complete-v0": 4.0,
    "kitchen-partial-v0": 4.0,
    "kitchen-mixed-v0": 4.0,
    "flow-ring-random-v0": 24.42,
    "flow-ring-controller-v0": 24.42,
    "flow-merge-random-v0": 330.03179,
    "flow-merge-controller-v0": 330.03179,
    "carla-lane-v0": 1023.5784385429523,
    "carla-town-v0": 2440.1772022247314,  # avg dataset score
    "bullet-halfcheetah-random-v0": 2381.6725,
    "bullet-halfcheetah-medium-v0": 2381.6725,
    "bullet-halfcheetah-expert-v0": 2381.6725,
    "bullet-halfcheetah-medium-expert-v0": 2381.6725,
    "bullet-halfcheetah-medium-replay-v0": 2381.6725,
    "bullet-hopper-random-v0": 1441.8059623430963,
    "bullet-hopper-medium-v0": 1441.8059623430963,
    "bullet-hopper-expert-v0": 1441.8059623430963,
    "bullet-hopper-medium-expert-v0": 1441.8059623430963,
    "bullet-hopper-medium-replay-v0": 1441.8059623430963,
    "bullet-ant-random-v0": 2650.495,
    "bullet-ant-medium-v0": 2650.495,
    "bullet-ant-expert-v0": 2650.495,
    "bullet-ant-medium-expert-v0": 2650.495,
    "bullet-ant-medium-replay-v0": 2650.495,
    "bullet-walker2d-random-v0": 1623.6476303317536,
    "bullet-walker2d-medium-v0": 1623.6476303317536,
    "bullet-walker2d-expert-v0": 1623.6476303317536,
    "bullet-walker2d-medium-expert-v0": 1623.6476303317536,
    "bullet-walker2d-medium-replay-v0": 1623.6476303317536,
    "bullet-maze2d-open-v0": 64.15,
    "bullet-maze2d-umaze-v0": 153.99,
    "bullet-maze2d-medium-v0": 238.05,
    "bullet-maze2d-large-v0": 285.92,
}



import gymnasium as gym
import minari
import numpy as np
import torch
from tensordict import TensorDict
from torch.utils.data import Dataset
from tqdm import tqdm


# %%
class MinariEpisodeDataset(Dataset):
    def __init__(self, dataset_name: str):
        self.dataset_name = dataset_name

        try:
            dataset = minari.load_dataset(dataset_name)
            self.dataset = dataset
            return
        except:
            self.dataset = None

    def create(self, env, n_episodes: int):
        assert self.dataset is None, "Dataset already exists"

        if not isinstance(env, minari.DataCollector):
            env = minari.DataCollector(env)

        for _ in tqdm(range(n_episodes), total=n_episodes, desc="Collecting data"):
            env.reset()
            done = False
            while not done:
                action = env.action_space.sample()  # <- use your policy here
                obs, rew, terminated, truncated, info = env.step(action)
                done = terminated or truncated
        dataset = env.create_dataset(self.dataset_name)
        self.dataset = dataset

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> minari.EpisodeData:
        return self.dataset[idx]

    def __iter__(self):
        return iter(self.dataset)

    def info(self):
        print(f"Observation space: {self.dataset.observation_space}")
        print(f"Action space: {self.dataset.action_space}")
        print(f"Total episodes: {self.dataset.total_episodes:,}")
        print(f"Total steps: {self.dataset.total_steps:,}")


class MinariTrajectoryDataset(Dataset):
    def __init__(self, minari_dataset: MinariEpisodeDataset, traj_len: int):
        self.minari_dataset = minari_dataset
        self.traj_len = traj_len
        self.book = self.build_book(self.minari_dataset, self.traj_len)

    @staticmethod
    def build_book(minari_dataset: MinariEpisodeDataset, traj_len: int):
        book = {}
        traj_id = 0

        for ep in minari_dataset:
            if ep.total_timesteps < traj_len:
                continue

            for start_idx in range(ep.total_timesteps - traj_len):
                book[traj_id] = (ep.id, start_idx, start_idx + traj_len)
                traj_id += 1

        return book

    def __len__(self) -> int:
        return len(self.book)

    def __getitem__(self, index: int | list[int]) -> TensorDict:
        match index:
            case int():
                return self._get_by_idx(index)
            case list():
                return torch.cat([self[i].unsqueeze(0) for i in index], dim=0)
            case slice():
                index = list(range(*index.indices(len(self))))
                return self[index]
            case np.ndarray():
                index = index.tolist()
                return self[index]
            case torch.Tensor():
                index = index.tolist()
                return self[index]
            case _:
                raise ValueError(f"Invalid index type: {type(index)}")

    def _get_by_idx(self, index: int) -> TensorDict:
        ep_id, start_idx, end_idx = self.book[index]
        ep = self.minari_dataset[ep_id]
        observations = ep.observations[start_idx:end_idx]
        actions = ep.actions[start_idx:end_idx]
        rewards = ep.rewards[start_idx:end_idx]
        terminations = ep.terminations[start_idx:end_idx]
        truncate = ep.truncations[start_idx:end_idx]

        return TensorDict(
            {
                "observations": torch.tensor(observations),
                "actions": torch.tensor(actions),
                "rewards": torch.tensor(rewards),
                "terminations": torch.tensor(terminations),
                "truncate": torch.tensor(truncate),
            },
            batch_size=[],
        )



import asyncio
import time

import gymnasium as gym
import minari
from torch.utils.data import DataLoader, Dataset

from .minari_ext import *


async def collect_episode(env, policy=None):
    """Collect a single episode using the specified environment and policy."""
    env.reset()
    done = False
    while not done:
        action = env.action_space.sample() if policy is None else policy()
        obs, rew, terminated, truncated, info = env.step(action)
        done = terminated or truncated


async def worker(env, queue, policy, start_time, print_every, total_episodes):
    collected = 0
    while True:
        episode_idx = await queue.get()
        if episode_idx is None:
            break
        await collect_episode(env, policy)
        collected += 1

        # Calculate and print ETA
        if collected % print_every == 0:
            elapsed_time = time.time() - start_time
            episodes_done = episode_idx + 1
            episodes_left = total_episodes - episodes_done
            eta = (
                (elapsed_time / episodes_done) * episodes_left
                if episodes_done > 0
                else float("inf")
            )
            eta_minutes, eta_seconds = divmod(eta, 60)
            print(
                f"Collected {episodes_done} episodes. ETA: {int(eta_minutes)}m {int(eta_seconds)}s"
            )


async def collect_samples_async(
    env_id, num_episodes, dataset_id, policy=None, num_envs=4, print_every=1000
):
    """Collect samples asynchronously using multiple environments."""
    start_time = time.time()

    # Initialize environments
    envs = [
        AsyncDataCollector(gym.make(env_id), record_infos=True) for _ in range(num_envs)
    ]

    # Create a queue for managing episode collection
    queue = asyncio.Queue()

    # Enqueue episode indices
    for i in range(num_episodes):
        queue.put_nowait(i)

    # Add termination signals to the queue for each worker
    for _ in range(num_envs):
        queue.put_nowait(None)

    # Start workers
    tasks = [
        worker(env, queue, policy, start_time, print_every, num_episodes)
        for env in envs
    ]
    await asyncio.gather(*tasks)

    # Create the dataset using the first environment
    dataset = envs[0].create_dataset(dataset_id)

    # Add data from other environments to the dataset
    for i, env in enumerate(envs[1:], start=2):
        try:
            env.add_to_dataset(dataset)
        except Exception as e:
            print(f"Error adding data from environment {i}: {e}")

    # Close all environments
    for env in envs:
        env.close()

    # Print total time taken
    total_time = time.time() - start_time
    total_minutes, total_seconds = divmod(total_time, 60)
    print(f"Total time taken: {int(total_minutes)}m {int(total_seconds)}s")

    return dataset


class OfflineRLDataset(Dataset):
    def __init__(
        self,
        dataset_id: str,
        env_id: str | None = None,
        num_episodes: int | None = None,
    ):

        try:
            dataset = minari.load_dataset(dataset_id)
            self.dataset = dataset
            return
        except:
            pass

        assert env_id is not None, "env_id must be provided if dataset_id is not found"
        assert (
            num_episodes is not None
        ), "num_episodes must be provided if dataset_id is not found"

        self.dataset_id = dataset_id
        self.env_id = env_id
        self.num_episodes = num_episodes
        self.dataset = None

    async def collect_episodes(self, num_envs: int = 1, print_every: int = 1000):
        if self.dataset:
            print("Dataset already exists")
            return

        await collect_samples_async(
            env_id=self.env_id,
            num_episodes=self.num_episodes,
            dataset_id=self.dataset_id,
            num_envs=num_envs,
            print_every=print_every,
        )
        self.dataset = minari.load_dataset(self.dataset_id)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int):
        return self.dataset[idx]
